from __future__ import annotations
from configparser import NoSectionError
import contextlib
import os
import git
import click
import copier
import ckan.plugins.toolkit as tk

TPL_DIR = os.path.join(os.path.dirname(__file__), "copier")
PROJECT_PREFIX = "ckanext-"


@click.group()
def ckanext():
    """Generate CKAN extension"""


@ckanext.command()
@click.option(
    "-o",
    "--output-dir",
    help="Location to put the generated template.",
    default=".",
)
@click.option(
    "-f",
    "--overwrite",
    help="Overwrite existing extension.",
    is_flag=True,
)
@click.argument("project", default=PROJECT_PREFIX)
@click.option("-d", "--use-defaults", is_flag=True)
def extended(output_dir: str, overwrite: bool, project: str, use_defaults: bool):
    """Generate empty extension files to expand CKAN."""
    template_loc = os.path.join(TPL_DIR, "extended")
    if not project.startswith(PROJECT_PREFIX):
        project = f"{PROJECT_PREFIX}{project}"


    if use_defaults and project == PROJECT_PREFIX:
        tk.error_shout("Project must be specified")
        raise click.Abort

    defaults = {"author": "", "author_email": "", "project": project}
    git_config = git.GitConfigParser()

    with contextlib.suppress(NoSectionError):
        defaults["author"] = git_config.get("user", "name")

    with contextlib.suppress(NoSectionError):
        defaults["author_email"] = git_config.get("user", "email")

    result = copier.run_copy(
        template_loc,
        output_dir,
        user_defaults=defaults,
        defaults=use_defaults,
        overwrite=overwrite,
        answers_file=".copier-answers.ctb-extended.yml",
        unsafe=True,
    )

    extension_path = os.path.realpath(
        os.path.join(output_dir, result.answers.combined["project"])
    )
    click.echo(f"Written: {extension_path}")



from __future__ import annotations

from typing import Any

from copier_templates_extensions import ContextHook


class ContextUpdater(ContextHook):
    def hook(self, context: dict[str, Any]) -> dict[str, Any]:
        patch: dict[str, Any] = {}
        return patch



from __future__ import annotations

import os
import uuid

import magic

import ckan.plugins as p
import ckan.plugins.toolkit as tk
from ckan.lib import uploader


class SafeUploadPlugin(p.SingletonPlugin):
    p.implements(p.IUploader, inherit=True)

    # IUploader

    def get_resource_uploader(self, data_dict):
        for plugin in p.PluginImplementations(p.IUploader):
            if isinstance(plugin, type(self)):
                continue
            return plugin.get_resource_uploader(data_dict)
        return None

    def get_uploader(self, upload_to, old_filename=None):
        return SafeUpload(upload_to, old_filename)


class SafeUpload(uploader.Upload):
    storage_path: str
    filename: str | None

    def update_data_dict(self, data_dict, url_field, file_field, clear_field):
        super().update_data_dict(data_dict, url_field, file_field, clear_field)
        if self.filename:
            self.verify_type()

            self.filename = str(uuid.uuid3(uuid.NAMESPACE_DNS, self.filename))
            self.filepath = os.path.join(self.storage_path, self.filename)
            data_dict[url_field] = self.filename

    def verify_type(self):
        if not self.filename:
            return

        actual = magic.from_buffer(self.upload_file.read(1024), mime=True)
        self.upload_file.seek(0, os.SEEK_SET)

        err = {self.file_field: [f"Unsupported upload type: {actual}"]}

        mimetypes = tk.aslist(
            tk.config.get(f"ckan.upload.{self.object_type}.mimetypes"),
        )
        if mimetypes and actual not in mimetypes:
            raise tk.ValidationError(err)

        type_ = actual.split("/")[0]
        types = tk.aslist(tk.config.get(f"ckan.upload.{self.object_type}.types"))
        if types and type_ not in types:
            raise tk.ValidationError(err)



from __future__ import annotations

from typing import Any

import ckan.plugins as p
from ckan import types


class PackageController(p.SingletonPlugin):
    """Customize dataset lifecycle."""

    p.implements(p.IPackageController, inherit=True)

    def after_dataset_show(
        self,
        context: types.Context,
        pkg_dict: dict[str, Any],
    ) -> None:
        """Add fake data."""
        pkg_dict["fake"] = 42

    def before_dataset_search(self, search_params: dict[str, Any]) -> dict[str, Any]:
        """Improve search filters."""
        return search_params



"""Tests for ckanext.{{ project_shortname }}.cli module.

Use `cli` fixture to invoke CLI commands.
"""

from __future__ import annotations

from typing import Any

from click.testing import CliRunner
from faker import Faker

from ckan import model

import ckanext.{{ project_shortname }}.cli as commands


class TestCommand:
    """Test `heh command` CLI command."""

    def test_output(self, cli: CliRunner, faker: Faker):
        """Command prints expected message."""
        # it's possible to execute `commands.command` directly, but group
        # contains initialization logic and `commands.command` will fail
        # without it
        #
        # Any parameter specified after the command must be passed inside a
        # second argument of `cli.invoke`. For example, `{{project_shortname}}
        # a b -C 1 --xxx=2` becomes `cli.invoke({{ project_shortname }}, ["a",
        # "b", "-C", "1", "--xxx=2"])`
        result = cli.invoke(commands.{{ project_shortname }}, ["command"])
        # `result.exit_code` contains `0` if command succeeded
        assert not result.exit_code
        # `result.output` contains normal output. `result.stderr` contains
        # output of an error
        assert result.output == "Hello, {{ project_shortname }}\n"

        # use fake data instead of hardcoded test values to discover
        # output-based problems
        name = faker.word()
        result = cli.invoke(commands.{{ project_shortname }}, ["command", name])
        assert result.output == f"Hello, {name}\n"

    def test_verbosity(self, cli: CliRunner):
        """Verbosity level appends exclamation marks."""
        result = cli.invoke(commands.{{ project_shortname }}, ["command", "-v"])
        assert result.output == "Hello, {{ project_shortname }}!\n"

        result = cli.invoke(commands.{{ project_shortname }}, ["command", "-vvv"])
        assert result.output == "Hello, {{ project_shortname }}!!!\n"


class TestCountUsers:
    """Test `heh count-users` CLI command."""

    def test_output(self, cli: CliRunner, faker: Faker, user_factory: Any):
        """Command counts number of users in DB."""
        result = cli.invoke(commands.{{ project_shortname }}, ["count-users"])
        count = model.User.count()
        assert result.output == f"\nResult: {count}!\n"

        # create a new user to verify that we are not receiving static result
        user_factory()

        result = cli.invoke(commands.{{ project_shortname }}, ["count-users"])
        count = model.User.count()
        assert result.output == f"\nResult: {count}!\n"



from __future__ import annotations

import click

import ckan.plugins.toolkit as tk
from ckan import model


@click.group(short_help="Extra tools for managing DB")
def db():
    pass


@db.command()
@click.option("-y", "--yes", is_flag=True, help="Confirm your intention.")
@click.option("-k", "--keep", multiple=True, help="Tables that should not be cleaned")
def clean(yes: bool, keep: tuple[str]):
    if not yes:
        tk.error_shout("This command will erase data from your portal's DB.")
        tk.error_shout("All the datasets, organizations and users will be removed")
        tk.error_shout("Run it with `--yes` flag if you know what you are doing.")
        raise click.Abort

    model.repo.session.remove()
    ## use raw connection for performance
    connection = model.repo.session.connection()
    tables = reversed(model.repo.metadata.sorted_tables)

    for table in tables:
        if table.name in keep:
            continue

        if not table.c:
            # skip `geometry_columns` or similar object from postgis
            continue

        if "alembic_version" in table.name:
            continue

        connection.execute('truncate "%s" cascade' % table.name)
    model.repo.session.commit()
    click.secho("Database table data deleted", fg="green")



from __future__ import annotations

from ckan import types
from ckan.logic.schema import validator_args


@validator_args
def get_sum(
    convert_int: types.Validator,
    not_empty: types.Validator,
) -> types.Schema:
    """Schema for {{ project_shortname }}_get_sum action."""
    return {
        "left": [not_empty, convert_int],
        "right": [not_empty, convert_int],
    }


@validator_args
def something_create(
    not_empty: types.Validator,
    unicode_safe: types.Validator,
    ignore_empty: types.Validator,
    convert_to_json_if_string: types.Validator,
    dict_only: types.Validator,
) -> types.Schema:
    """Schema for {{ project_shortname }}_something_create action."""
    return {
        "hello": [not_empty, unicode_safe],
        "world": [not_empty, unicode_safe],
        "plugin_data": [ignore_empty, convert_to_json_if_string, dict_only],
    }



"""Template helpers of the {{ project_shortname }} plugin.

All non-private functions defined here are registered inside `tk.h` collection.
"""

from __future__ import annotations


def {{project_shortname}}_hello() -> str:
    """Greet the user.

    Returns:
        greeting with the plugin name.
    """
    return "Hello, {{project_shortname}}!"



"""Tests for ckanext.{{ project_shortname }}.logic.action."""

import pytest
from faker import Faker

from ckan import model
from ckan.tests.helpers import call_action

from ckanext.{{ project_shortname }}.model import Something


@pytest.mark.ckan_config("ckan.plugins", "{{project_shortname}}")
@pytest.mark.usefixtures("with_plugins")
def test_get_sum():
    """Sum is correct."""
    result = call_action("{{project_shortname}}_get_sum", left=10, right=30)
    assert result["sum"] == 40


@pytest.mark.ckan_config("ckan.plugins", "{{project_shortname}}")
@pytest.mark.usefixtures("with_plugins")
def test_something_create(faker: Faker):
    """Something is created."""
    hello = faker.word()
    world = faker.word()

    # action tests use `call_action` to skip authorization check. All auth
    # functions are checked separately.
    result = call_action(
        "{{project_shortname}}_something_create",
        hello=hello,
        world=world,
    )

    # if validation covered by schema, do not test in inside action
    # test. Basically, test the code, not the process.
    assert result["hello"] == hello
    assert result["world"] == world

    smth = model.Session.get(Something, result["id"])
    assert smth.hello == hello



from __future__ import annotations

import pytest

import ckan.plugins.core as pcore
from ckan.tests.pytest_ckan import fixtures

if not hasattr(pcore, "unload_non_system_plugins"):

    @pytest.fixture()
    def with_plugins(ckan_config):
        """Override CKAN `with_plugins` unloading **all** plugins, even if they
        are not listed in `ckan.plugins`.

        """
        pcore.load_all()
        yield

        system_plugins = pcore.find_system_plugins()
        plugins_to_unload = [
            p
            for p in reversed(pcore._PLUGINS)  # type: ignore
            if p not in system_plugins
        ]
        pcore.unload(*plugins_to_unload)


if fixtures.migrate_db_for._pytestfixturefunction.scope == "function":

    @pytest.fixture(scope="session")
    def migrate_db_for():
        """Override CKAN `migrate_db_for` using `session` scope for fixture."""

        from ckan.cli.db import _run_migrations

        def runner(plugin, version="head", forward=True):
            assert plugin, "Cannot apply migrations of unknown plugin"
            _run_migrations(plugin, version, forward)

        return runner


if not hasattr(fixtures, "reset_db_once") and not hasattr(fixtures, "non_clean_db"):

    @pytest.fixture(scope="session")
    def reset_db_once(reset_db):
        """Internal fixture that cleans DB only the first time it's used."""
        reset_db()

    @pytest.fixture(scope="function")
    def non_clean_db(reset_db_once):
        """Guarantees that DB is initialized.

        This fixture either initializes DB if it hasn't been done yet or does
        nothing otherwise. If there is some data in DB, it stays intact. If your
        tests need empty database, use `clean_db` instead, which is much slower,
        but guarantees that there are no data left from the previous test session.

        Example::

            @pytest.mark.usefixtures("non_clean_db")
            def test_example():
                assert factories.User()

        """
        from ckan import model

        model.repo.init_db()


if not hasattr(fixtures, "reset_redis") and not hasattr(fixtures, "clean_redis"):

    @pytest.fixture(scope="session")
    def reset_redis():
        """Callable for removing all keys from Redis.
        Accepts redis key-pattern for narrowing down the list of items to
        remove. By default removes everything.
        This fixture removes all the records from Redis on call::
            def test_redis_is_empty(reset_redis):
                redis = connect_to_redis()
                redis.set("test", "test")
                reset_redis()
                assert not redis.get("test")
        If only specific records require removal, pass a pattern to the fixture::
            def test_redis_is_empty(reset_redis):
                redis = connect_to_redis()
                redis.set("AAA-1", 1)
                redis.set("AAA-2", 2)
                redis.set("BBB-3", 3)
                reset_redis("AAA-*")
                assert not redis.get("AAA-1")
                assert not redis.get("AAA-2")
                assert redis.get("BBB-3") is not None
        """

        def cleaner(pattern: str = "*") -> int:
            """Remove keys matching pattern.
            Return number of removed records.
            """
            from ckan.lib.redis import connect_to_redis

            conn = connect_to_redis()
            keys = conn.keys(pattern)
            if keys:
                return conn.delete(*keys)
            return 0

        return cleaner

    @pytest.fixture()
    def clean_redis(reset_redis):
        """Remove all keys from Redis.
        This fixture removes all the records from Redis::
            @pytest.mark.usefixtures("clean_redis")
            def test_redis_is_empty():
                assert redis.keys("*") == []
        If test requires presence of some initial data in redis, make sure that
        data producer applied **after** ``clean_redis``::
            @pytest.mark.usefixtures(
                "clean_redis",
                "fixture_that_adds_xxx_key_to_redis"
            )
            def test_redis_has_one_record():
                assert redis.keys("*") == [b"xxx"]
        """
        reset_redis()


if not hasattr(fixtures, "app_with_session"):

    @pytest.fixture()
    def app_with_session(make_app):
        from flask.sessions import SecureCookieSessionInterface

        app = make_app()
        app.flask_app.session_interface = SecureCookieSessionInterface()
        return app



from __future__ import annotations

from typing import Any

import ckan.plugins.toolkit as tk
from ckan import model
from ckan.logic import validate
from ckan.types import Context

from ckanext.{{project_shortname}}.model import Something

from . import schema


@tk.side_effect_free
@validate(schema.get_sum)
def {{project_shortname}}_get_sum(context: Context, data_dict: dict[str, Any]):
    """Produce a sum of left and right.

    Args:
        left (int): firt argument
        right (int): second argument

    Returns:
        operation details
    """
    tk.check_access("{{project_shortname}}_get_sum", context, data_dict)

    return {
        "left": data_dict["left"],
        "right": data_dict["right"],
        "sum": data_dict["left"] + data_dict["right"],
    }


@validate(schema.something_create)
def {{project_shortname}}_something_create(context: Context, data_dict: dict[str, Any]):
    """Create something object.

    Args:
        hello (str): aliquam erat volutpat
        world: (str): nullam tempus
        plugin_data (dict[str, Any], optional): aliquam feugiat tellus ut neque

    Returns:
        details of the new something object
    """
    tk.check_access("{{project_shortname}}_something_create", context, data_dict)

    smth = Something(
        hello=data_dict["hello"],
        world=data_dict["world"],
        plugin_data=data_dict.get("plugin_data", {}),
    )
    model.Session.add(smth)
    model.Session.commit()

    return smth.dictize(context)
